from __future__ import annotations
from typing import Dict, List, Optional


import cim.data_profile as cim

def add_to_catalog(obj: object, catalog: Dict) -> Dict:
    if obj.mRID == None:
        raise ValueError('Object must contain an mRID')
    if obj.mRID not in catalog:
        catalog[obj.mRID] = obj
    return catalog

def add_to_typed_catalog(obj: object, typed_catalog: Dict) -> Dict:
    if type(obj) not in typed_catalog:
        typed_catalog[type(obj)] = {}
    if obj.mRID not in typed_catalog[type(obj)]:
        typed_catalog[type(obj)][obj.mRID] = obj
    return typed_catalog


def get_all_by_type(typed_catalog: Dict, obj_type: type) -> List[object]:
    objects = typed_catalog.get(obj_type) if typed_catalog.get(obj_type) is not None else []
    for obj in objects:
        if obj.mRID not in self.is_loaded:
            load_all_attributes(obj)


from typing import List
from dataclasses import dataclass, field

import cim.data_profile as cim

def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?length ?bch ?r ?x ?gch ?b0ch ?r0 ?x0 ?g0ch 
        ?PerLengthImpedance ?WireSpacingInfo
        (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        (group_concat(distinct ?ACLineSegmentPhase; separator=';') as ?ACLineSegmentPhases)  
        WHERE {          
          ?eq r:type cim:ACLineSegment.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        ?eq cim:ConductingEquipment.BaseVoltage ?bv.
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage.

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminal

        OPTIONAL {?meas cim:Measurement.Terminal ?t.
          ?meas cim:IdentifiedObject.mRID ?Measurement}

        OPTIONAL {?eq cim:Conductor.length ?length.}
        
        OPTIONAL {?eq cim:ACLineSegment.bch ?bch.}
        OPTIONAL {?eq cim:ACLineSegment.r ?r0.}
        OPTIONAL {?eq cim:ACLineSegment.x ?x0.}
        OPTIONAL {?eq cim:ACLineSegment.g ?r0.}
        OPTIONAL {?eq cim:ACLineSegment.b0ch ?b0.}
        OPTIONAL {?eq cim:ACLineSegment.r0 ?r0.}
        OPTIONAL {?eq cim:ACLineSegment.x0 ?x0.}
        OPTIONAL {?eq cim:ACLineSegment.g0ch ?b0.}

        OPTIONAL {?eq cim:ACLineSegment.PerLengthImpedance ?pli.
                  ?pli cim:IdentifiedObject.mRID ?PerLengthImpedance.}
        OPTIONAL {?eq cim:ACLineSegment.WireSpacingInfo ?wsi.
                  ?wsi cim:IdentifiedObject.mRID ?WireSpacingInfo.}
        OPTIONAL {?aclsp cim:ACLineSegmentPhase.ACLineSegment ?eq.
                  ?aclsp cim:IdentifiedObject.mRID ?ACLineSegmentPhase.}

        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?length ?bch ?r ?x ?gch ?b0ch ?r0 ?x0 ?g0ch 
                ?PerLengthImpedance ?WireSpacingInfo 
        ORDER by  ?name
        """
    return query_message


from __future__ import annotations
from typing import Dict, List, Optional
import traceback
from cim.models.model_parsers import add_to_catalog, add_to_typed_catalog
import cim.data_profile as cim

def query_parser(typed_catalog:Dict, class_name:str, mRID:str, query:List, attribute:str, separator:str) -> object | str:
#     try:
    value = query[attribute]['value']

    #if attribute is CIM class, then build CIM objects. otherwise assign to obj_list
#         print(attribute)
    if attribute in cim.__all__:
        value = value.split(separator)
        result = build_cim_object(typed_catalog, attribute, value)
    else:
        result = value
    setattr(typed_catalog[class_name][mRID], attribute, result)
#     except Exception:
#         traceback.print_exc()


def query_list_parser(typed_catalog:Dict, class_name:type, mRID:str, query:List, attribute:str, attribute_class:str, separator:str):
    value = query[attribute]['value']
    values = value.split(separator)
    #if attribute is CIM class, then build CIM objects. otherwise assign to obj_list
    if attribute_class in cim.__all__:
        obj_list = build_cim_object(typed_catalog, attribute_class, values)
    else:
        obj_list = values
    #set attribute of queried object to list parsed from query results
    setattr(typed_catalog[class_name][mRID], attribute, obj_list)
    

def build_cim_object(typed_catalog:Dict, class_name:str, mRID_list:List[str]) -> List(object):
    cls = class_name
    #add class to typed_catalog if not already defined
    if class_name not in typed_catalog.keys():
        typed_catalog[class_name] = {}
    obj_list = []
    for mRID in mRID_list:
        #add object to typed_catalog if not already included
        if mRID in typed_catalog[class_name].keys():
            obj = typed_catalog[class_name][mRID]
        else:
            obj = eval(f"cim.{cls}(mRID='{mRID}')")
            add_to_typed_catalog(obj, typed_catalog)
        obj_list.append(obj)
    return obj_list



from __future__ import annotations

from cim.data_profile import Feeder


def format_terminal_all_attributes_by_feeder(feeder_id: str | Feeder) -> str:
    query_message = """
        # list all the terminals
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX c:  <http://iec.ch/TC57/CIM100#>
        SELECT ?tid ?tname ?fdrid ?cnid ?tpid ?coneqid ?txfid ?regcntl ?seq
        {
          VALUES ?fdrid {%s} 
          ?fdr c:IdentifiedObject.mRID ?fdrid.

          ?t c:Terminal.ConnectivityNode ?cn. 
          bind(strafter(str(?cn),"#") as ?cnid).

          ?t c:Terminal.ConductingEquipment ?coneq.
          bind(strafter(str(?coneq),"#") as ?coneqid).

          OPTIONAL {?t c:Terminal.TransformerEnd ?txf.
          bind(strafter(str(?txf),"#") as ?txfid).} 

          OPTIONAL {?t c:Terminal.RegulatingControl ?regcntl}


          ?t c:IdentifiedObject.name ?tname.
          bind(strafter(str(?t),"#") as ?tid).

          ?cn c:ConnectivityNode.TopologicalNode ?tp.
          bind(strafter(str(?tp),"#") as ?tpid).

          ?t c:ACDCTerminal.sequenceNumber ?seq.


        }
        GROUP BY  ?tid ?tname ?fdrid ?cnid ?tpid ?coneqid ?txfid ?regcntl ?seq
        ORDER by  ?fdrid 
    """ % feeder_id

    return query_message



from typing import List
from dataclasses import dataclass, field

import cim.data_profile as cim

def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?Location ?p ?q ?phase ?EnergyConsumer
        WHERE {          
          ?eq r:type cim:EnergyConsumerPhase.
          #VALUES ?fdrid {"%s"} #feeder_id association missing!
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        OPTIONAL {?eq cim:EnergyConsumerPhase.p ?p.}
        OPTIONAL {?eq cim:EnergyConsumerPhase.q ?q.}
        OPTIONAL {?eq cim:EnergyConsumerPhase.phase ?phase.}

        OPTIONAL {?eq cim:EnergyConsumerPhase.EnergyConsumer ?ec.
                  ?ec cim:IdentifiedObject.mRID ?EnergyConsumer.}
        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?p ?q ?phase ?EnergyConsumer
        ORDER by  ?name
        """
    return query_message


from __future__ import annotations
from typing import Dict, List, Optional
from SPARQLWrapper import SPARQLWrapper, JSON, POST
import re
from cim.loaders import ConnectionInterface, ConnectionParameters, Parameter, QueryResponse
from cim.loaders.blazegraph.query_parsers import query_parser, query_list_parser, build_cim_object
import cim.loaders.sparql as sparql
import cim.data_profile as cim



class BlazegraphConnection(ConnectionInterface):
    sparql_obj: Optional[SPARQLWrapper] = None

    def connect(self):
        if not self.sparql_obj:
            url = self.connection_params.parameters[0].value
            self.sparql_obj = SPARQLWrapper(url)
            self.sparql_obj.setReturnFormat(JSON)

    def disconnect(self):
        self.sparql_obj = None
        
    def execute(self, query_message: str) -> QueryResponse:
        self.connect()
        self.sparql_obj.setQuery(query_message)
        self.sparql_obj.setMethod(POST)
        query_output = self.sparql_obj.query().convert()
        return query_output

        
    def create_default_instances(self, feeder_mrid: str | cim.Feeder, mrid_list: List[str]) -> List[object]:
        """ 
        Creates empty CIM objects with the correct class type with mRID and name fields populated based on 
        a list of mRID strings.
        Args:
            feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
            mrid_list (list[str]): A list of object mRID strings to be converted into CIM objects
        Returns:
            object_list: A list of CIM object instances
        """
        #generate correct sparql message using create_default.py
        sparql_message = sparql.create_default_sparql(feeder_mrid, mrid_list)
        #execute sparql query
        query_output = self.execute(sparql_message)
        # parse query results and add new CIM objects to list
        object_list = [] 
        for result in query_output['results']['bindings']:
           # print(result)
            cls = result['class']['value']
            mRID = result['mRID']['value']
            name = result['name']['value']
            try:
                object_list.append(eval(f"cim.{cls}(mRID='{mRID}', name = '{name}')"))
                print(cls)
            except:
                print('warning: object class missing from data profile:', cls)
        return object_list
    
    
          
    def get_all_attributes(self, feeder_mrid: str | cim.Feeder, typed_catalog: dict[type, dict[str, object]], cim_class: type):
        """ Populates all available attribute fields of CIM objects in the typed catalog of a specified CIM class. 
        Objects are stored in memory, so no values are returned.
        Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
        cim_class (type): The CIM class type (e.g. cim:ACLineSegment)
        Returns:
        none
        """
        mrid_list = list(typed_catalog[cim_class].keys()) #get all object mRIDs in switch area
        #generate SPARQL message from correct loaders>sparql python script based on class name
        sparql_message = eval(f"sparql.{cim_class.__name__}SPARQL.get_all_attributes('{feeder_mrid}', {mrid_list})") 
        #execute sparql query
        query_output = self.execute(sparql_message)
#         print(query_output)

#         attribute_list = query_output['head']['vars'] #list of attributes received in query response
        for result in query_output['results']['bindings']: #iterate through rows of response
            attribute_list = result.keys()
            mRID = result['mRID']['value']
            name = result['name']['value']
            for attribute in attribute_list: 
                try: #check if attribute is in data profile
                    attribute_type = cim_class.__dataclass_fields__[attribute].type
                except:
                    #replace with warning message
                    print('warning: attribute ', attribute, ' missing from ', cim_class)
                    
                if 'List' in attribute_type: #check if attribute is association to a list of class objects
                    if '\'' in attribute_type: #handling inconsistent '' marks in data profile
                        at_cls = re.match(r'List\[\'(.*)\']',attribute_type)
                        attribute_class = at_cls.group(1)
                    else:        
                        at_cls = re.match(r'List\[(.*)]',attribute_type)
                        attribute_class = at_cls.group(1)
#                     print(attribute_class)
                    # pass query response of associated objects to list parser
                    query_list_parser(typed_catalog, cim_class, mRID, result, attribute, attribute_class, ';')

                else: #otherwise assign query response
#                     print(attribute)
#                     print(result)
                    query_parser(typed_catalog, cim_class, mRID, result, attribute, ';')
                    

  


from __future__ import annotations

import atexit
import os
from typing import Dict, List

import cim.data_profile as cim
from cim.loaders import ConnectionInterface, ConnectionParameters, Parameter, QueryResponse
from cim.models import add_to_catalog, add_to_typed_catalog
from gridappsd import GridAPPSD
from SPARQLWrapper import SPARQLWrapper, JSON, POST
import cim.loaders.sparql as sparql


# os.environ["GRIDAPPSD_ADDRESS"] = "gridappsd"
# os.environ["GRIDAPPSD_PORT"] = "61613"
os.environ['GRIDAPPSD_APPLICATION_ID'] = 'gridappsd-cim-profile'
os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STARTED'
os.environ['GRIDAPPSD_USER'] = 'app_user'
os.environ['GRIDAPPSD_PASSWORD'] = '1234App'

def query_string_parser(obj_dict: Dict, mRID: str, query: List, key: str, separator):
    try:
        value = query[key]['value']
        if separator in value:
            value = value.split(separator)
        setattr(obj_dict[mRID], key, value)
    except:
        []

def query_class_parser(typed_catalog, mRID, query, class_name, separator):
    try: 
        cls = class_name
        value = query[class_name]['value']
        if separator in value:
            values = value.split(separator)
            obj = []
            for mrid in values:
                obj.append(eval(f"cim.{cls}(mRID='{mrid}')"))
        else:
            obj = eval(f"cim.{cls}(mRID='{value}')")
        setattr(typed_catalog[mRID], class_name, obj)
    except:
        []
    return typed_catalog

def query_list_parser(obj_dict: Dict, mRID: str, query: List, key, separator):

    try:
        value = query[key]['value'].split(separator)
        if getattr(obj_dict[mRID], key) == None:
            setattr(obj_dict[mRID], key, value)
        else:
            setattr(obj_dict[mRID], key, getattr(obj_dict[mRID], key) + value)
    except:
        []


class GridappsdConnection(ConnectionInterface):
    __gapps__ = None

    def connect(self):
        if self.__gapps__ is None:
            self.__gapps__ = GridAPPSD()

    def disconnect(self):
        if self.__gapps__ is not None:
            self.__gapps__.disconnect()

    def load_attributes(self, obj: object):
        if isinstance(obj, cim.Terminal):
            # load terminal stuff here
            pass

    def create_default_instances(self, feeder_mrid: str | cim.Feeder, mrid_list: List[str]) -> List[object]:
        """ Creates an empty CIM object with the correct class type with mRID and name fields populated
        Args:
            feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
            mrid_list (list[str]): A list of object mRID strings to be converted into CIM objects
        Returns:
            objects: A list of CIM object instances
        """
        sparql_message = sparql.create_default_sparql(feeder_mrid, mrid_list, False)
        objects = self.query_object_builder(sparql_message)
        return objects
    
    def get_all_attributes(self, feeder_mrid, typed_catalog, cim_class):
        if cim_class in typed_catalog:
            mrid_list = list(typed_catalog[cim_class].keys())
            sparql_message = eval(f"sparql.{cim_class.__name__}SPARQL.get_all_attributes('{feeder_mrid}', {mrid_list})")
            self.query_attribute_builder(sparql_message, typed_catalog, cim_class)
    
    def query_object_builder(self, query_message):
        self.connect()
        request = {"requestType": "QUERY",
                   "resultFormat": "JSON",
                   "queryString": query_message}
        query_output = self.execute(topic='goss.gridappsd.process.request.data.powergridmodel',message=request, timeout=5)
        
        objects = []
        for result in query_output['data']['results']['bindings']:
            cls = result['class']['value']
            mRID = result['mRID']['value']
            name = result['name']['value']
            try:
                objects.append(eval(f"cim.{cls}(mRID='{mRID}', name = '{name}')"))
            except:
                print('warning: object class missing:', cls)
        return objects

    def execute(self, **kwargs) -> QueryResponse:
        for x in ('topic', 'message', 'timeout'):
            if x not in kwargs:
                raise ValueError(f"Parameter {x} required")

        response = self.__gapps__.get_response(**kwargs)
        return response
    
    def query_attribute_builder(self, query_message, typed_catalog, cim_class):
        self.connect()
        request = {"requestType": "QUERY",
                   "resultFormat": "JSON",
                   "queryString": query_message}
        query_output = self.execute(topic='goss.gridappsd.process.request.data.powergridmodel',message=request, timeout=5)
        attribute_list = query_output['data']['head']['vars']
        for result in query_output['data']['results']['bindings']:
            mRID = result['mRID']['value']
            name = result['name']['value']
            for attribute in attribute_list:
                if attribute == 'Measurements' or attribute == 'Terminals':
                    query_list_parser(typed_catalog[cim_class], mRID, result, attribute, ';')
                elif attribute in cim.__all__:
                    query_class_parser(typed_catalog[cim_class], mRID, result, attribute, ';')
                    try:
                        add_to_typed_catalog(getattr(typed_catalog[cim_class][mRID], attribute), typed_catalog)
                    except:
                        []
                else:
                    query_string_parser(typed_catalog[cim_class], mRID, result, attribute, ';')
#             query_list_parser(typed_catalog[cim_class], mRID, result, 'Measurements', ';')
#             query_list_parser(typed_catalog[cim_class], mRID, result, 'Terminals', ';')


def get_topology_response(feeder_mrid: str) -> QueryResponse:
    assert feeder_mrid is not None

    gapps = GridAPPSD()
    topic = "goss.gridappsd.request.data.topology"
    message = {
        "requestType": "GET_SWITCH_AREAS",
        "modelID": feeder_mrid,
        "resultFormat": "JSON"
    }

    topo_response = gapps.get_response(topic=topic, message=message, timeout=30)
    return topo_response

# Close gridappsd connection when exiting program.
#atexit.register(lambda: __gapps__.disconnect())



from typing import List
from dataclasses import dataclass, field

import cim.data_profile as cim


def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?constructionKind ?diameterOverInsulation ?diameterOverJacket ?diameterOverScreen
         ?isStrandFill ?nominalTemperature ?sheathAsNeutral ?shieldMaterial
        WHERE {          
          ?eq r:type cim:CableInfo.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
         OPTIONAL {?eq cim:CableInfo.constructionKind ?constructionKind.} #CableConstructionKind
         OPTIONAL {?eq cim:CableInfo.diameterOverCore ?diameterOverCore.} #Length
         OPTIONAL {?eq cim:CableInfo.diameterOverInsulation ?diameterOverInsulation.} #Length
         OPTIONAL {?eq cim:CableInfo.diameterOverJacket ?diameterOverJacket.} #Length
         OPTIONAL {?eq cim:CableInfo.diameterOverScreen ?diameterOverScreen} #Length
         OPTIONAL {?eq cim:CableInfo.isStrandFill ?isStrandFill } #Boolean,
         OPTIONAL {?eq cim:CableInfo.nominalTemperature ?nominalTemperature.} #Temperature,
         OPTIONAL {?eq cim:CableInfo.outerJacketKind ?outerJacketKind.} #CableOuterJacketKind,
         OPTIONAL {?eq cim:CableInfo.sheathAsNeutral ?sheathAsNeutral.} #Boolean,
         OPTIONAL {?eq cim:CableInfo.shieldMaterial ?shieldMaterial.} #CableShieldMaterialKind,




        }
        ORDER by  ?name
        """
    return query_message


from cim.loaders.sparql.create_default import create_default_sparql

import cim.loaders.sparql.ac_dc_terminal as ACDCTerminalSPARQL
import cim.loaders.sparql.ac_line_segment as ACLineSegmentSPARQL
import cim.loaders.sparql.ac_line_segment_phase as ACLineSegmentPhaseSPARQL
import cim.loaders.sparql.asset as AssetSPARQL
# import cim.loaders.sparql.asset_container as AssetContainerSPARQL
import cim.loaders.sparql.asset_info as AssetInfoSPARQL
import cim.loaders.sparql.base_voltage as BaseVoltageSPARQL
import cim.loaders.sparql.breaker as BreakerSPARQL
# import cim.loaders.sparql.bus_bar_section as BusBarSectionSPARQL
# import cim.loaders.sparql.bus_bar_section_info as BusBarSectionInfoSPARQL
import cim.loaders.sparql.cable_info as CableInfoSPARQL

import cim.loaders.sparql.energy_consumer as EnergyConsumerSPARQL
import cim.loaders.sparql.energy_consumer_phase as EnergyConsumerPhaseSPARQL


import cim.loaders.sparql.linear_shunt_compensator as LinearShuntCompensatorSPARQL



import cim.loaders.sparql.power_transformer as PowerTransformerSPARQL
# import cim.loaders.sparql.transformer_tank as TransformerTankSPARQL

import cim.loaders.sparql.wire_position as WirePositionSPARQL
import cim.loaders.sparql.wire_spacing_info as WireSpacingInfoSPARQL


from __future__ import annotations
from typing import List
from dataclasses import dataclass, field
import json

import cim.data_profile as cim
from cim.loaders import ConnectionInterface
from cim.models.model_parsers import add_to_catalog, add_to_typed_catalog




@dataclass
class SecondaryArea:
    area_id: str
    connection: ConnectionInterface
    distribution_transformer: dict[str, object] = field(default_factory=dict)
    addressable_equipment: dict[str, object] = field(default_factory=dict)
    unaddressable_equipment: dict[str, object] = field(default_factory=dict)
    connectivity_nodes: dict[str, object] = field(default_factory=dict)
    secondary_areas: list[object] = field(default_factory=list)
    typed_catalog: dict[type, dict[str, object]] = field(default_factory=dict)

    # Initialize empty CIM objects for all equipment in secondary area
    def initialize_secondary_area(self, switch_msg: dict):
        self.feeder_mrid = self.area_id.split('.')[0]  # get feeder mRID from area id

        addr_equip = self.connection.create_default_instances(self.feeder_mrid, switch_msg['addressable_equipment'])
        for obj in addr_equip:
            add_to_catalog(obj, self.addressable_equipment)
            add_to_typed_catalog(obj, self.typed_catalog)
        unaddr_equip = self.connection.create_default_instances(self.feeder_mrid, switch_msg['unaddressable_equipment'])
        for obj in unaddr_equip:
            add_to_catalog(obj, self.unaddressable_equipment)
            add_to_typed_catalog(obj, self.typed_catalog)
        conn_nodes = self.connection.create_default_instances(self.feeder_mrid, switch_msg['connectivity_node'])
        for obj in conn_nodes:
            add_to_catalog(obj, self.connectivity_nodes)
            add_to_typed_catalog(obj, self.typed_catalog)
        xfmr = self.connection.create_default_instances(self.feeder_mrid, switch_msg['distribution_transformer'])
        for obj in xfmr:
            add_to_catalog(obj, self.distribution_transformer)
            add_to_typed_catalog(obj, self.typed_catalog)
            
    def get_all_attributes(self, cim_class):
        if cim_class in self.typed_catalog:
            self.connection.get_all_attributes(self.feeder_mrid, self.typed_catalog, cim_class)
#             return self.__dumps__(cim_class)
            
    def __dumps__(self, cim_class):
        mrid_list = list(self.typed_catalog[cim_class].keys())
        attribute_list = list(cim_class().__dict__.keys())
        json_dump = {}

        for mrid in mrid_list:
            json_dump[mrid] = {}
            for attribute in attribute_list:
                value = getattr(self.typed_catalog[cim_class][mrid], attribute)
                if type(value) in [str, list]:
                    json_dump[mrid][attribute] = value
                elif value is None:
                    json_dump[mrid][attribute] = ''
                else:
                    json_dump[mrid][attribute] = value.__dict__

        return json.dumps(json_dump)


from typing import List
from dataclasses import dataclass, field

import cim.data_profile as cim

def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?p ?q ?customerCount ?grounded ?phaseConnection ?LoadResponse
            (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
            (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
            (group_concat(distinct ?EnergyConsumerPhase; separator=";") as ?EnergyConsumerPhase)
        WHERE {          
          ?eq r:type cim:EnergyConsumer.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.

        ?eq cim:ConductingEquipment.BaseVoltage ?bv.
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage.

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminal

        OPTIONAL {?meas cim:Measurement.Terminal ?t.
        ?meas cim:IdentifiedObject.mRID ?Measurement}

        OPTIONAL {?eq cim:EnergyConsumer.p ?p.}
        OPTIONAL {?eq cim:EnergyConsumer.q ?q.}
        OPTIONAL {?eq cim:EnergyConsumer.customerCount ?customerCount.}
        OPTIONAL {?eq cim:EnergyConsumer.grounded ?grounded.}
        
        OPTIONAL {?eq cim:EnergyConsumer.phaseConnection ?phs.
                  bind(strafter(str(?phs),"PhaseShuntConnectionKind.") as ?phaseConnection) }
        OPTIONAL {?eq cim:EnergyComsumer.LoadResponse ?lr.
                  ?lr cim:IdentifiedObject.mRID ?LoadResponse}
        OPTIONAL {?ecp cim:EnergyConsumerPhase.EnergyConsumer ?eq.
                  ?ecp cim:IdentifiedObject.mRID ?EnergyConsumerPhase.}

        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?p ?q ?customerCount ?grounded ?phaseConnection ?LoadResponse
        ORDER by  ?name
        """
    return query_message


from typing import List
from dataclasses import dataclass, field

import cim.data_profile as cim

def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?Location ?vectorGroup
        (group_concat(distinct ?PowerTransformerEnd; separator=";") as ?PowerTransformerEnd) 
        (group_concat(distinct ?TransformerTank; separator=';') as ?TransformerTanks)
        WHERE {          
          ?eq r:type cim:PowerTransformer.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        OPTIONAL {?eq cim:PowerTransformer.vectorGroup ?vectorGroup.}

        OPTIONAL {?tank cim:TransformerTank.PowerTransformer ?eq.
                  ?tank cim:IdentifiedObject.mRID ?TransformerTank.}

        OPTIONAL {?end cim:PowerTransformerEnd.PowerTransformer ?eq.
                  ?end cim:IdentifiedObject.mRID ?PowerTransformerEnd.}
        }
        GROUP by ?mRID ?name ?Location ?vectorGroup
        ORDER by ?name
        """
    return query_message